.statusBar {
  height: fit-content;
  background-color: #333;

  display: flex;
  flex-direction: row;
  justify-content: space-between;

  color: white;

  padding-left: 10px;
}

.statusBarRight {
  display: flex;
  flex-direction: row;
  justify-content: space-between;
}

.spacerMiddle {
  flex-grow: 1;
}

.spacerLeft,
.spacerRight {
  width: 1rem;
}

.socketError {
  color: red;
}

.socketStatus svg {
  display: inline-block;
  width: 100%;
  height: 100%;
  max-width: 16px;
  max-height: 16px;
  color: white;
  transform: translateY(2px);
}
