.resizeHandle {
  display: flex;
  align-items: center;
  justify-content: center;
  height: 100%;
  aspect-ratio: 1 / 1;
  cursor: nwse-resize;
}

body.resizing {
  pointer-events: none;
  user-select: none;
}

body.resizing * {
  pointer-events: none !important;
}

.resizeIcon {
  width: 100%;
  height: 100%;
  max-width: 16px;
  max-height: 16px;
  color: white;
}
